# distutils: language = c++
# cython: language_level = 3
cimport numpy as np
import numpy as np

np.import_array()

from runtime.runtime cimport BatchMailBox, NullResult, Scheduler, WaitResult
from libc.stdio cimport printf
from libc.stdlib cimport malloc, free
from libc.math cimport fmin, fmax, fabs
from libcpp.vector cimport vector
from openmp cimport omp_get_max_threads

from cython.parallel import prange

## Types declaration
ctypedef int I_t
ctypedef double D_t

cdef lock Scheduler scheduler
cdef D_t INF = 1e38

# NOTE: The following extern definition is used to interface
# std::nth_element, a robust partitioning algorithm, in Cython 
cdef extern from *:
    """
    #include <algorithm>
    #include <iostream>
    template<class D, class I>
    class IndexComparator {
    private:
        const D *data;
        I split_dim, n_features;
    public:
        IndexComparator(const D *data, const I &split_dim, const I &n_features):
            data(data), split_dim(split_dim), n_features(n_features) {}
        bool operator()(const I &a, const I &b) const {
            D a_value = data[a * n_features + split_dim];
            D b_value = data[b * n_features + split_dim];
            return a_value == b_value ? a < b : a_value < b_value;
        }
    };
    template<class D, class I>
    void partition_node_indices(
        const D *data,
        I *node_indices,
        I start,
        I mid,
        I end,
        const I &split_dim,
        const I &n_features) {
        IndexComparator<D, I> index_comparator(data, split_dim, n_features);

        std::nth_element(
            node_indices + start,
            node_indices + mid,
            node_indices + end,
            index_comparator
        );

    }
    """
    void partition_node_indices[D, I](
                D *data_ptr,
                I *node_indices,
                I start,
                I mid,
                I end,
                I split_dim,
                I n_features
    ) nogil except +


cdef cypclass Counter activable:
    """ A simple Counter.
    
    This can be useful for synchronisation for the caller after
    triggering the actors logic as it wait for the value of
    the Coutner to reach a given one before moving on.
    """

    I_t _n

    __init__(self):
        self._n = 0
        self._active_result_class = WaitResult.construct
        self._active_queue_class = consume BatchMailBox(scheduler)

    void add(self, I_t value):
        self._n += value

    void reset(self):
        self._n = 0

    I_t value(self):
        return self._n


cdef cypclass Container activable:
    """ A simple wrapper of an array.
    
    The wrapped array is passed by the initial caller which then
    trigger the actors logic modifying it.

    The initial caller can wait for specific number of update
    before proceeding.
    """

    I_t * _array
    I_t _size
    I_t _n_updates


    __init__(self, I_t *array, I_t size):
        self._array = array
        self._size = size
        self._active_result_class = WaitResult.construct
        self._active_queue_class = consume BatchMailBox(scheduler)
        self._n_updates = 0

    void update(self, I_t idx, I_t value):
        self._n_updates += 1
        self._array[idx] = value

    int n_updates(self):
        return self._n_updates


cdef inline void dual_swap(D_t* darr, I_t* iarr, I_t i1, I_t i2) nogil:
    """swap the values at inex i1 and i2 of both darr and iarr"""
    cdef D_t dtmp = darr[i1]
    darr[i1] = darr[i2]
    darr[i2] = dtmp

    cdef I_t itmp = iarr[i1]
    iarr[i1] = iarr[i2]
    iarr[i2] = itmp


cdef void _simultaneous_sort(
    D_t* dist,
    I_t* idx,
    I_t size
) nogil:
    """
    Perform a recursive quicksort on the dist array, simultaneously
    performing the same swaps on the idx array.  The equivalent in
    numpy (though quite a bit slower) is

    def simultaneous_sort(dist, idx):
        i = np.argsort(dist)
        return dist[i], idx[i]
    """
    cdef I_t pivot_idx, i, store_idx
    cdef D_t pivot_val

    # in the small-array case, do things efficiently
    if size <= 1:
        pass
    elif size == 2:
        if dist[0] > dist[1]:
            dual_swap(dist, idx, 0, 1)
    elif size == 3:
        if dist[0] > dist[1]:
            dual_swap(dist, idx, 0, 1)
        if dist[1] > dist[2]:
            dual_swap(dist, idx, 1, 2)
            if dist[0] > dist[1]:
                dual_swap(dist, idx, 0, 1)
    else:
        # Determine the pivot using the median-of-three rule.
        # The smallest of the three is moved to the beginning of the array,
        # the middle (the pivot value) is moved to the end, and the largest
        # is moved to the pivot index.
        pivot_idx = size // 2
        if dist[0] > dist[size - 1]:
            dual_swap(dist, idx, 0, size - 1)
        if dist[size - 1] > dist[pivot_idx]:
            dual_swap(dist, idx, size - 1, pivot_idx)
            if dist[0] > dist[size - 1]:
                dual_swap(dist, idx, 0, size - 1)
        pivot_val = dist[size - 1]

        # partition indices about pivot.  At the end of this operation,
        # pivot_idx will contain the pivot value, everything to the left
        # will be smaller, and everything to the right will be larger.
        store_idx = 0
        for i in range(size - 1):
            if dist[i] < pivot_val:
                dual_swap(dist, idx, i, store_idx)
                store_idx += 1
        dual_swap(dist, idx, store_idx, size - 1)
        pivot_idx = store_idx

        # recursively sort each side of the pivot
        if pivot_idx > 1:
            _simultaneous_sort(dist, idx, pivot_idx)
        if pivot_idx + 2 < size:
            _simultaneous_sort(dist + pivot_idx + 1,
                                    idx + pivot_idx + 1,
                                    size - pivot_idx - 1)


cdef cypclass NeighborsHeap activable:
    """A max-heap structure to keep track of distances/indices of neighbors

    This implements an efficient pre-allocated set of fixed-size heaps
    for chasing neighbors, holding both an index and a distance.
    When any row of the heap is full, adding an additional point will push
    the furthest point off the heap.

    Taken and adapted from:
    https://github.com/scikit-learn/scikit-learn/blob/e4bb9fa86b0df873ad750b6d59090843d9d23d50/sklearn/neighbors/_binary_tree.pxi#L513

    The initial caller is responsible for providing the array of indices
    which will be modified in place by the actors logic.

    n_pushes and is_sorted can be used by the initial caller to know
    when to pursue.

    Parameters
    ----------
    n_pts : int
        the number of heaps to use
    n_nbrs : int
        the size of each heap.
    """
    D_t *_distances
    I_t *_indices

    I_t _n_pts
    I_t _n_nbrs
    I_t _n_pushes
    bint _sorted

    __init__(self, I_t * indices, I_t n_pts, I_t n_nbrs):
        cdef I_t i
        self._n_pts = n_pts
        self._n_nbrs = n_nbrs
        self._active_result_class = WaitResult.construct
        self._active_queue_class = consume BatchMailBox(scheduler)
        self._distances = <D_t *> malloc(n_pts * n_nbrs * sizeof(D_t))
        self._indices = indices
        self._n_pushes = 0
        self._sorted = False

        # We can't use memset here
        for i in range(n_pts * n_nbrs):
            self._distances[i] = INF

    void __dealloc__(self):
        free(self._distances)

    void increment_counter(self, I_t val):
        self._n_pushes += val

    void push(self, I_t row, D_t val, I_t i_val):
        """push (val, i_val) into the given row"""
        cdef:
            I_t i, left_child_idx, right_child_idx, swap_idx

            # Getting the heap to use
            I_t *indices = self._indices + row * self._n_nbrs
            D_t *distances = self._distances + row * self._n_nbrs

        self._n_pushes += 1

        # check if val should be in heap
        if val > distances[0]:
            return

        # insert val at position zero
        distances[0] = val
        indices[0] = i_val

        # descend the heap, swapping values until the max heap criterion is met
        i = 0
        while True:
            left_child_idx = 2 * i + 1
            right_child_idx = left_child_idx + 1

            if left_child_idx >= self._n_nbrs:
                break
            elif right_child_idx >= self._n_nbrs:
                if distances[left_child_idx] > val:
                    swap_idx = left_child_idx
                else:
                    break
            elif distances[left_child_idx] >= distances[right_child_idx]:
                if val < distances[left_child_idx]:
                    swap_idx = left_child_idx
                else:
                    break
            else:
                if val < distances[right_child_idx]:
                    swap_idx = right_child_idx
                else:
                    break

            distances[i] = distances[swap_idx]
            indices[i] = indices[swap_idx]

            i = swap_idx

        distances[i] = val
        indices[i] = i_val

    void sort(self):
        # NOTE: Ideally we could sort results in parallel, but
        # OpenMP threadpool and this runtime's aren't working
        # nicely together (using prange here would create OpenMP
        # threads within workers, causing unexpected behavior.)
        cdef I_t row

        for row in range(self._n_pts):
            # We use a function here to be able to recurse
            _simultaneous_sort(
                self._distances + row * self._n_nbrs,
                self._indices + row * self._n_nbrs,
                self._n_nbrs)

        self._sorted = True

    int n_pushes(self):
        return self._n_pushes

    int is_larger(self, I_t row, D_t val):
        # TODO: support for double via promises?
        return 1 if val > self._distances[row * self._n_nbrs] else 0

    int is_sorted(self):
        # TODO: is there a support for returning bool via
        # promises? As of now returning a int for making
        # use of getIntResult
        return 1 if self._sorted else 0


cdef cypclass Node activable:
    """A KDTree Node
    
    Node delegate tasks to their children Nodes.

    Some Nodes are set as Leaves when they are associated
    to ``leaf_size`` or less points.
    
    Leafs are terminal Nodes and do not have children.
    """

    # Reference to the head of the allocated arrays
    # data gets not modified via _data_ptr 
    D_t * _data_ptr
    I_t * _indices_ptr

    # The point the Node split on
    D_t * _point
    I_t _n_dims
    I_t _dim

    # Portion of _indices covered by the Node is:
    # _indices_ptr[_start:_end]
    I_t _start
    I_t _end

    bint _is_leaf

    active Node _top
    active Node _left
    active Node _right

    vector[D_t] _lower_bounds
    vector[D_t] _upper_bounds

    __init__(self):
        self._active_result_class = WaitResult.construct
        self._active_queue_class = consume BatchMailBox(scheduler)
        self._left = NULL
        self._right = NULL
        self._is_leaf = False


    # We use this to allow using actors for initialisation
    # because __init__ can't be reified. 
    void build_node(
            self,
            active Node top,
            D_t * data_ptr,
            I_t * indices_ptr,
            I_t leaf_size,
            I_t n_dims,
            I_t dim,
            I_t start,
            I_t end,
            active Counter counter,
    ):
        cdef:
            I_t k
            I_t next_dim = (dim + 1) % n_dims
            I_t mid = (start + end) // 2
            D_t * point

        self._top = top

        self._data_ptr = data_ptr
        self._indices_ptr = indices_ptr

        self._dim = dim
        self._n_dims = n_dims
        self._start = start
        self._end = end

        self._upper_bounds.assign(self._n_dims, -INF)
        self._lower_bounds.assign(self._n_dims, INF)

        if (end - start <= leaf_size):
            self._is_leaf = True

            # Compute the actual data range to be able to prune then
            for i in range(self._start + 1, self._end):
                point = self._data_ptr + self._indices_ptr[i] * self._n_dims
                for k in range(self._n_dims):
                    self._lower_bounds[k] = fmin(self._lower_bounds[k], point[k])
                    self._upper_bounds[k] = fmax(self._upper_bounds[k], point[k])

            if self._top is not NULL and self._top is not self:
                # Notifying the top Node of changes
                for k in range(self._n_dims):
                    self._top._notify_lower_bound(NULL, k, self._lower_bounds[k])
                    self._top._notify_upper_bound(NULL, k, self._upper_bounds[k])

            # Adding to the global counter the number
            # of samples the leaf is responsible of
            counter.add(NULL, end - start)
            return

        # We partition the samples in two nodes on a given dimension,
        # with the middle point as a pivot
        partition_node_indices(data_ptr, indices_ptr, start, mid, end, dim, n_dims)
        self._point = data_ptr + mid

        self._left = consume Node()
        self._right = consume Node()

        # Recursing on both partitions
        self._left.build_node(NULL, consume self,
                              data_ptr, indices_ptr,
                              leaf_size, n_dims, next_dim,
                              start, mid, counter)
        self._right.build_node(NULL, consume self,
                               data_ptr, indices_ptr,
                               leaf_size, n_dims, next_dim,
                               mid, end, counter)

    void _notify_lower_bound(self, I_t dim, D_t value):
        if self._lower_bounds[dim] > value:
            self._lower_bounds[dim] = value
            if self._top is not NULL and self._top is not self:
                self._top._notify_lower_bound(NULL, dim, value)

    void _notify_upper_bound(self,  I_t dim, D_t value):
        if self._upper_bounds[dim] < value:
            self._upper_bounds[dim] = value
            if self._top is not NULL and self._top is not self:
                self._top._notify_upper_bound(NULL, dim, value)

    void query(self,
               D_t * query_points,
               I_t i,
               active NeighborsHeap heaps,
    ):
        cdef:
            I_t k
            D_t dist = 0.0
            D_t * point = query_points + i * self._n_dims
            D_t d, d_lo, d_hi

        # Pruning strategy: we don't explore the node
        # if it is outside the bounds
        for k in range(self._n_dims):
            d_lo = self._lower_bounds[k] - point[k]
            d_hi = point[k] - self._upper_bounds[k]
            # Here we  use the fact that:
            #       x + abs(x) = 2 * max(x, 0)
            d = (d_lo + fabs(d_lo)) + (d_hi + fabs(d_hi))
            dist += 0.5 * d * d

        if heaps.is_larger(NULL, i, dist).getIntResult():
            # We dont't explore this node here
            heaps.increment_counter(NULL, self._end - self._start)
            return

        if self._is_leaf:
            # Computing all the euclideans distances here
            for j in range(self._start, self._end):
                dist = 0
                for k in range(self._n_dims):
                    d = (
                        point[k] -
                        self._data_ptr[self._indices_ptr[j] * self._n_dims + k]
                        )
                    dist += d * d
                
                # The heap is doing the smart work of keeping
                # the closest points for each query point i
                heaps.push(NULL, i, dist, self._indices_ptr[j])
            return

        self._left.query(NULL, query_points, i, heaps)
        self._right.query(NULL, query_points, i, heaps)


cdef cypclass KDTree:
    """A KDTree based on asynchronous and parallel computations.

    It is similar to :class:`sklearn.neighbours.KDTree` but it is
    using Cython+ features for asynchronous computations.

    Asynchronuous and parallel computations can be used when
    constructing the nodes (see above) and when querying (WIP).

    This relies on a Cython+ runtime using actors.
    """

    I_t _n # number of examples
    I_t _d # number of dimensions / features
    I_t _leaf_size # maximum number of vectors at leaf

    I_t _n_leafs

    active Node _root

    D_t *_data_ptr
    I_t *_indices_ptr

    __init__(self,
             np.ndarray X,
             I_t leaf_size,
             ):
        cdef I_t i
        cdef I_t n = X.shape[0]
        cdef I_t d = X.shape[1]
        cdef I_t initialised = 0
        # Accessing _SC_NPROCESSORS_ONLN does not return the 
        # effective number of threads which were assigned to
        # this task using `taskset(1)` for instance. 
        # 
        # This OpenMP API is a workable way to access it.
        cdef I_t num_workers = omp_get_max_threads()

        self._n = n
        self._d = d
        self._leaf_size = leaf_size

        self._data_ptr = <D_t *> X.data
        self._indices_ptr = <I_t *> malloc(n * sizeof(I_t))
        for i in range(n):
            self._indices_ptr[i] = i

        # Recurvisely building the tree here
        global scheduler
        scheduler = Scheduler(num_workers)
        cdef active Counter counter = consume Counter()

        self._root = consume Node()
        if self._root is NULL:
            printf("Error consuming node\n")

        # When object are activated (set as Actors), methods
        # are reified. When using those reified methods
        # a new argument is prepredend for a predicate,
        # which we aren't using using here, hence the extra NULL.
        #
        # Also using this separate method allowing using actors
        # because __init__ can't be reified.
        self._root.build_node(NULL, consume self._root,
        # TODO: there's a problem of types here (^) at C++
        # compile time. Probably, we need the new type system
        # before pursuing.
                              self._data_ptr,
                              self._indices_ptr,
                              self._leaf_size, n_dims=self._d,
                              dim=0, start=0, end=self._n,
                              counter=counter)

        # Waiting for the tree construction to end
        # Somewhat similar to a thread barrier
        while(initialised < self._n):
            initialised = counter.value(NULL).getIntResult()

        counter.reset(NULL)
        

    void __dealloc__(self):
        scheduler.finish()
        free(self._indices_ptr)


    void query(self,
               np.ndarray query_points,  # IN
               np.ndarray closests,      # OUT
               ):
        cdef:
            I_t completed_queries = 0
            I_t i
            I_t n_query = query_points.shape[0]
            I_t n_neighbors = closests.shape[1]
            I_t total_n_pushes = n_query * self._n

            active NeighborsHeap heaps

        heaps = consume NeighborsHeap(<I_t *> closests.data,
                                      n_query,
                                      n_neighbors)

        for i in range(n_query):
            self._root.query(NULL, <D_t *> query_points.data, i, heaps)

        while(completed_queries < total_n_pushes):
            completed_queries = heaps.n_pushes(NULL).getIntResult()

        heaps.sort(NULL)
        
        while not(heaps.is_sorted(NULL).getIntResult()):
            pass


cdef public int main() nogil:
    # Entry point for the compiled binary file
    printf("empty public int main() nogil:")
    return 0
